#include "lfpaillier.h"

void djcs_encrypt_with_length(djcs_public_key *pk,hcs_random *hr, mpz_t ctx, mpz_t plain, unsigned long s)
{
	pk->s = s;
	djcs_encrypt(pk,hr,ctx,plain);
}

void djcs_decrypt_with_length(djcs_private_key *sk, mpz_t plain, mpz_t ctx, unsigned long s)
{
	sk->s = s;
	djcs_decrypt(sk,plain,ctx);
}

void djcs_hom_add_ptx(djcs_public_key *pk,mpz_t res, mpz_t op1, mpz_t plain,unsigned long s)
{
	pk->s = s;
	djcs_ep_add(pk,res,op1,plain);
}

void djcs_hom_add(djcs_public_key *pk,mpz_t res, mpz_t op1, mpz_t op2,unsigned long s)
{
	pk->s = s;
	djcs_ee_add(pk,res,op1,op2);
}

void djcs_hom_mul(djcs_public_key *pk,mpz_t res, mpz_t op1, mpz_t plain,unsigned long s)
{
	pk->s = s;
	djcs_ep_mul(pk,res,op1,plain);
}

void djcs_cleanup(djcs_public_key *pk, djcs_private_key *sk, hcs_random *hr)
{
	djcs_clear_public_key(pk);
	free(pk);
	djcs_clear_private_key(sk);
	free(sk);
	hcs_free_random(hr);
}

void invert_bit(mpz_t flipped, djcs_public_key *pk, mpz_t bit, unsigned long s)
{
	mpz_invert(flipped,bit,pk->n[s]);
	mpz_mul(flipped,flipped,pk->g);
	mpz_mod(flipped,flipped,pk->n[s]);
	/*mpz_invert(flipped_selector,selector_bit,pk->n[s]);
	mpz_set(tmp,pk->n[s-1]);
	mpz_sub_ui(tmp,tmp,1);
	djcs_hom_mul(pk,flipped_selector,selector_bit,tmp,s);
	mpz_set_ui(tmp,1);
	djcs_hom_add_ptx(pk,flipped_selector,flipped_selector,tmp,s);
	*/
}

void oblivious_selection(mpz_t dest,djcs_public_key *pk, mpz_t el1, mpz_t el2, mpz_t selector_bit, unsigned long s)
{
	mpz_t flipped_selector,tmp;
	mpz_inits(flipped_selector,tmp,NULL);
	invert_bit(flipped_selector,pk,selector_bit,s);
	djcs_hom_mul(pk,tmp,flipped_selector,el1,s);
	djcs_hom_mul(pk,dest,selector_bit,el2,s);
	djcs_hom_add(pk,dest,dest,tmp,s);
	mpz_clears(flipped_selector,tmp,NULL);
}

void oblivious_selection(mpz_t dest,djcs_public_key *pk, mpz_t el1, mpz_t el2, mpz_t flipped_selector, mpz_t selector_bit, unsigned long s)
{
	mpz_t tmp;
	mpz_init(tmp);
	//invert_bit(flipped_selector,pk,selector_bit,s);
	djcs_hom_mul(pk,tmp,flipped_selector,el1,s);
	djcs_hom_mul(pk,dest,selector_bit,el2,s);
	djcs_hom_add(pk,dest,dest,tmp,s);
	mpz_clear(tmp);
}


void homomorphic_dot_product(mpz_t dest, djcs_public_key *pk, mpz_t * results, mpz_t* trapdoor, unsigned long s, unsigned int b)
{
	mpz_t tmp;
	mpz_init(tmp);
	djcs_hom_mul(pk,dest,trapdoor[0],results[0],s);
	for(int i=1;i<b;i++)
	{
		djcs_hom_mul(pk,tmp,trapdoor[i],results[i],s);
		djcs_hom_add(pk,dest,dest,tmp,s);
	}
}
