#include "lipmaa.h"
#include "build_bwt.h"

void pir_trapdoor(mpz_t* trapdoor, unsigned int index,unsigned int l,unsigned int b,djcs_public_key *pk,hcs_random *hr)
{
	//mpz_t* trapdoor = (mpz_t *) malloc(l*sizeof(mpz_t));
	mpz_t one, zero;
	mpz_inits(one,zero,NULL);
	for(int i=0;i<l;i++)
	{
		//mpz_init(trapdoor[i]);
		mpz_set_ui(one,1);
		mpz_set_ui(zero,0);
		unsigned int digit = index % b;
		for(int j=0;j<b;j++)
		{
			if(digit == j)
				djcs_encrypt_with_length(pk,hr,trapdoor[b*i+j],one,i+1);
			else
				djcs_encrypt_with_length(pk,hr,trapdoor[b*i+j],zero,i+1);
		}
		index = index / b;
	}
	mpz_clears(one,zero,NULL);
	//return trapdoor;
}

unsigned int dump_trapdoor(mpz_t *trapdoor, unsigned int l, unsigned int b, unsigned int modulus_size,const char*filename)
{
	unsigned long num_digits;
	unsigned char*el;
	FILE *fp = fopen(filename,"wb");
	mpz_t tmp;
	mpz_init(tmp);
	unsigned int count = 0;
	for(int i=0;i<l;i++)
	{
		num_digits = modulus_size*(i+1)/8;
		for(int j=0;j<b;j++)
		{
			mpz_set(tmp,trapdoor[i*b+j]);
			el = mpz_to_byte(tmp,num_digits);
			fwrite(el,sizeof(unsigned long),num_digits,fp);
			count += sizeof(unsigned long)*num_digits;
			free(el);
		}
	}
	fclose(fp);
	return count;
}

void verify_trapdoor(mpz_t *trapdoor,djcs_private_key *sk,unsigned long l,unsigned int b)
{
	mpz_t plain;
	mpz_init(plain);
	for(int i=0;i<l;i++)
	{
		for(int j=0;j<b;j++)
		{
			djcs_decrypt_with_length(sk,plain,trapdoor[b*i+j],i+1);
			gmp_printf("%Zd ",plain);
		}
		printf("\t");
	}
	printf("\n");
}

void recursive_pir(mpz_t res,djcs_public_key *pk,mpz_t *trapdoor,mpz_t* complement_trapdoor, mpz_t *A,unsigned long l,unsigned int b, unsigned int begin,unsigned int end,unsigned int len)
{
	if (begin >= len)
	{
		mpz_set_ui(res,1);
		return;
	}	
	if (end - begin == 0)
	{
		mpz_set(res,A[begin]);
		//printf("%d %d \n",begin,l);
		//fflush(stdout);
		return;
	}
	//printf("begin: %d end: %d depth: %d ", begin, end,l);
	//fflush(stdout);
	unsigned int bucket_size = (end - begin) / b;
	//mpz_t* results = (mpz_t*) malloc(sizeof(mpz_t)*b);
	mpz_t el;
	mpz_init(el);
	//mpz_inits(el1,el2,NULL);
	recursive_pir(el,pk,trapdoor,complement_trapdoor,A,l-1,b,begin,begin+bucket_size,len);
	begin += bucket_size +1;
	mpz_t tmp;
	mpz_init(tmp);
	//djcs_hom_mul(pk,res,trapdoor[(l-1)*b],el,l);
	/*for(int i=1;i<b;i++)
	{
	}*/
	mpz_powm(res,trapdoor[(l-1)*b],el,pk->n[l]);
	for(int i=1;i<b;i++)
	{
		//mpz_init(results[i]);
	//mpz_inits(el1,el2,NULL);
		if(begin >= len)
			break;
		recursive_pir(el,pk,trapdoor,complement_trapdoor,A,l-1,b,begin,begin+bucket_size,len);
		begin += bucket_size +1;
		//djcs_hom_mul(pk,tmp,trapdoor[(l-1)*b+i],el,l);
		mpz_powm(tmp,trapdoor[(l-1)*b+i],el,pk->n[l]);
		//djcs_hom_add(pk,res,res,tmp,l);
		mpz_mul(res,res,tmp);
		mpz_mod(res,res,pk->n[l]);
	}
	//homomorphic_dot_product(res,pk,results,&(trapdoor[(l-1)*b]),l,b);
		//recursive_pir(el2,pk,trapdoor,complement_trapdoor,A,l-1,middle+1,end,len);
	//oblivious_selection(res,pk,el1,el2,complement_trapdoor[l-1],trapdoor[l-1],l);
	//oblivious_selection(res,pk,el1,el2,trapdoor[l-1],l);
	//mpz_clears(el1,el2,NULL);
	mpz_clears(el,tmp,NULL);
	//for(int i=0;i<b;i++)
	//	mpz_clear(results[i]);
	//free(results);
}

void pir_search(mpz_t res,djcs_public_key *pk,mpz_t *trapdoor,unsigned long l,mpz_t *A,unsigned int len,unsigned int b)
{
	/*mpz_t* complement_trapdoor = (mpz_t *) malloc(l*sizeof(mpz_t));
	for(int i=0;i<l;i++)
	{
		mpz_init(complement_trapdoor[i]);
		invert_bit(complement_trapdoor[i],pk,trapdoor[i],i+1);
	}*/
	unsigned int power = 1;
	for(int i=0;i<l;i++)
		power *= b;
	recursive_pir(res,pk,trapdoor,NULL,A,l,b,0,power-1,len);
	/*for(int i=0;i<l;i++)
		mpz_clear(complement_trapdoor[i]);
	free(complement_trapdoor);*/
}

void pir_retrieve(unsigned char* el, djcs_private_key *sk,unsigned char* enc_key, unsigned long l, unsigned long index, int entry_size, mpz_t ciphertext)
{
	for(;l>0;l--)
	{
		djcs_decrypt_with_length(sk,ciphertext,ciphertext,l);
	}
	unsigned int num_digits = (entry_size)/8;
	unsigned char* el_ctx = mpz_to_byte(ciphertext,(entry_size)%8 == 0 ? num_digits : num_digits +1 );
	int len;
	EVP_CIPHER_CTX* ctx = EVP_CIPHER_CTX_new();
	if(!ctx)
	{
		printf("Wrong ctx initialization \n");
		exit(-1);
	}
	if(1 != EVP_DecryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,enc_key,el_ctx) )
	{
		printf("decryption initialization error \n");
		exit(-1);
	}
	if(1 != EVP_DecryptUpdate(ctx,(unsigned char *)el,&len,el_ctx+BLOCK_SIZE,entry_size-BLOCK_SIZE) )
	{
		printf("decryption failed \n");
		exit(-1);
	}
	if(1 != EVP_DecryptFinal_ex(ctx, ((unsigned char *) el) + len, &len))
	{
		printf("final decryption failed \n");
		exit(-1);
	}	
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
}

