#include "build_bwt.h"
#include "lipmaa.h"
#include <vector>
#include <string>
#include <thread>


const int alphabet_size = 5;
typedef fasta_nucleotide charset;
typedef std::uint32_t Integer;
int b;
const int sa_sample_rate = 28;

void byte_to_mpz(unsigned char* bytes,unsigned int len_bytes,mpz_t dest)
{
	mpz_set_ui(dest,0);
	unsigned long *el = (unsigned long*) bytes;
	mpz_add_ui(dest,dest,*el);
	for(int i=sizeof(unsigned long);i<len_bytes;i+=sizeof(unsigned long))
	{
		el = (unsigned long*) &(bytes[i]);
		mpz_mul_2exp(dest,dest,64);
		mpz_add_ui(dest,dest,*el);
	}
}

unsigned char* mpz_to_byte(mpz_t op, unsigned int num_digits)
{
	unsigned long* dest = (unsigned long*) malloc(sizeof(unsigned long)*num_digits);
	CHECK_MALLOC(dest);
	mpz_t tmp;
	mpz_inits(tmp, NULL);
	for(int i=num_digits;i>0;i--)
	{
		mpz_fdiv_r_2exp(tmp,op,64);
		mpz_fdiv_q_2exp(op,op,64);
		dest[i-1] = mpz_get_ui(tmp); 	
	}
	mpz_clear(tmp);
	return (unsigned char*) dest;
}


int build_sampled_bwt(unsigned char* enc_key,mpz_t *A,std::uint8_t* ptx_array, int entry_size, int array_size)
{	
	unsigned char iv[BLOCK_SIZE];
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}

	EVP_CIPHER_CTX *ctx;
	ctx = EVP_CIPHER_CTX_new();
	if(!ctx)
	{
		printf("Wrong ctx initialization \n");
		exit(-1);
	}
	if(1 != EVP_EncryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,enc_key,NULL) )
	{
		printf("encryption initialization error \n");
		exit(-1);
	}
	std::uint8_t *el_ctx = (std::uint8_t*) malloc((entry_size+1)*BLOCK_SIZE);
	CHECK_MALLOC(el_ctx);
	int ctx_len,len,j=0;
	for(unsigned int i=0;i<array_size;i+=entry_size)
	{
		fread(iv, sizeof(unsigned char), BLOCK_SIZE, fp);
		if(1 != EVP_EncryptInit_ex(ctx,NULL,NULL,NULL,iv))
		{
			printf("encryption iv setup failed \n");
			exit(-1);
		}
		memcpy(el_ctx,iv,BLOCK_SIZE);
		ctx_len = BLOCK_SIZE;
		if(1 != EVP_EncryptUpdate(ctx,el_ctx+ctx_len,&len,&(ptx_array[i]),entry_size))
		{
			printf("encryption failed \n");
			exit(-1);
		}
		ctx_len += len; 
		if(1 != EVP_EncryptFinal_ex(ctx,el_ctx+ctx_len,&len))
		{
			printf("padding failed \n");
			exit(-1);
		}
		ctx_len += len;
		mpz_init(A[j]);
		byte_to_mpz(el_ctx,ctx_len,A[j]);	
		//gmp_printf(" element %d is %Zd \n",j,A[j]);
		j++;
	}
	fclose(fp);
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
	return ctx_len;

}


template <typename Int>
int build_suffix_array(unsigned char* enc_key,mpz_t *A,Int* ptx_array,int sample_rate, int N)
{
	unsigned char iv[BLOCK_SIZE];
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}

	EVP_CIPHER_CTX *ctx;
	ctx = EVP_CIPHER_CTX_new();
	if(!ctx)
	{
		printf("Wrong ctx initialization \n");
		exit(-1);
	}
	if(1 != EVP_EncryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,enc_key,NULL) )
	{
		printf("encryption initialization error \n");
		exit(-1);
	}
	unsigned char *element,*el_ctx = (unsigned char *) malloc(sizeof(Int)*sample_rate + 2*BLOCK_SIZE);
	CHECK_MALLOC(el_ctx);
	int ctx_len,len,j=0;
	for(unsigned long i=0;i<N;i+=sample_rate)
	{
		//unsigned char * element = malloc(sizeof(unsigned long));
		//memcpy(element,&i,sizeof(unsigned long));
		element = (unsigned char *) &(ptx_array[i]);
		fread(iv, sizeof(unsigned char), BLOCK_SIZE, fp);
		if(1 != EVP_EncryptInit_ex(ctx,NULL,NULL,NULL,iv))
		{
			printf("encryption iv setup failed \n");
			exit(-1);
		}
		memcpy(el_ctx,iv,BLOCK_SIZE);
		ctx_len = BLOCK_SIZE;
		if(1 != EVP_EncryptUpdate(ctx,el_ctx+ctx_len,&len,element,sizeof(Int)*sample_rate))
		{
			printf("encryption failed \n");
			exit(-1);
		}
		ctx_len += len; 
		if(1 != EVP_EncryptFinal_ex(ctx,el_ctx+ctx_len,&len))
		{
			printf("padding failed \n");
			exit(-1);
		}
		ctx_len += len;
		mpz_init(A[j]);
		byte_to_mpz(el_ctx,ctx_len,A[j]);
		j++;
	}
	fclose(fp);
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
	return ctx_len;
}


void build_occurrences_array(unsigned long * occ_array, unsigned long* ch_count,charset *bwt,unsigned long length,int alphabet_size,int terminator)
{
	for(int i=0; i < alphabet_size +1;i++)
		ch_count[i] = 0;
	for(int i=0;i<length+1;i++)
	{
		if(i == terminator)
			ch_count[0]++;
		else
			ch_count[(Integer) bwt[i] + 1]++;
		for(int j=0;j<alphabet_size+1;j++)
			occ_array[i+j*(length+1)] = ch_count[j];
	}
	for(int j=1;j < alphabet_size + 1; j++)
		ch_count[j] += ch_count[j-1]; 
	/*for(int j=alphabet_size;j > 0; j--)
		ch_count[j] = ch_count[j-1];
	ch_count[0] = 0;*/
	for(int i=1;i < alphabet_size+1;i++)
		for(int j=0;j<length+1;j++)
			occ_array[i*(length+1)+j] += ch_count[i-1];
}

unsigned long* query_ptx(charset* substring, int m, Integer* ch_count,std::uint8_t* BWT, dataset* data, Integer* SA, unsigned long* num_occ)
{
	/*Integer ch_index = ((Integer) substring[m-1]); //+1 because the first row of the bwt is due to the terminator
	unsigned long f = ch_count[ch_index] - 1; //-1 needs to be added because indexes of the array starts from 0
	unsigned long g = ch_count[ch_index+1] - 1;
	for(int i=m-2;i>=0;i--)
	{
		ch_index = ((Integer) substring[i]) + 1; //+1 because the first row of the bwt is due to the terminator
		f = ch_index*length + f;
		f = BWT[f] - 1;
		g = ch_index*length + g;
		g = BWT[g] - 1;
	}*/
	Integer start, end;
	vbwt_query<Integer,charset>(BWT,ch_count,data->sample_rate,data->sample_size,data->no_bits,alphabet_size,substring,m,&start,&end);
	
	*num_occ = end-start+1;
	if(*num_occ != 0)
	{
		unsigned long* occ = new unsigned long[*num_occ];
		for(int i=start;i<end+1;i++)
			occ[i-start] = SA[i];
		return occ;
	}
	return NULL;
}

void query(charset* substring, int m, Integer* ch_count,dataset* data, int max_depth, key_tuple* key, unsigned long* num_occ,unsigned long **priv_occ)
{
	Integer ch_index = ((Integer) substring[m-1]);
	unsigned long f = ch_count[ch_index]; //-1 needs to be added because indexes of the array starts from 0
	unsigned long g = ch_count[ch_index+1] - 1;
	unsigned long offset;
	std::uint32_t* blob = (std::uint32_t*) malloc(data->sample_size);
	CHECK_MALLOC(blob);
	std::uint16_t* ch;
	mpz_t * trapdoor = (mpz_t*) malloc(sizeof(mpz_t)*max_depth*b);
	CHECK_MALLOC(trapdoor);
	for(int i=0;i<max_depth*b;i++)
		mpz_init(trapdoor[i]);
	mpz_t res;
	mpz_init(res);
	clock_t begin, end;
	//double trap_time = 0, search_time= 0, ret_time = 0;
	//begin = clock();
	long start, theend;
	struct timeval timecheck;
	gettimeofday(&timecheck, NULL);
        start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
	for(int i=m-2;i>=0;i--)
	{
		//printf("start is % d end is %d \n",f,g);
		ch_index = ((Integer) substring[i]);
		offset = f % data->sample_rate;
		f = f / data->sample_rate;
		//begin = clock();
		pir_trapdoor(trapdoor,f,data->bwt_depth,b,key->pk,key->hr);
		//end = clock();
		//trap_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		dump_trapdoor(trapdoor,data->bwt_depth,b,1024,"trapdoor");
		//begin = clock();
		pir_search(res,key->pk,trapdoor,data->bwt_depth,data->BWT,data->bwt_len,b);
		//end = clock();
		//search_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		//begin = clock();
		pir_retrieve((unsigned char*) blob,key->sk,key->enc_key,data->bwt_depth,f,data->bwt_entry_size,res);
		//end = clock();
		ch = (std::uint16_t *) (blob + alphabet_size);
		f = ch_count[ch_index];
		f += blob[ch_index] + partial_rank(ch,offset,ch_index,data->no_bits);
		//ret_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		offset = g % data->sample_rate + 1;
		g = g / data->sample_rate;
		pir_trapdoor(trapdoor,g,data->bwt_depth,b,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->bwt_depth,data->BWT,data->bwt_len,b);
		pir_retrieve((unsigned char*) blob,key->sk,key->enc_key,data->bwt_depth,g,data->bwt_entry_size,res);
		ch = (std::uint16_t *) (blob + alphabet_size);
		g = ch_count[ch_index];
		g += blob[ch_index] + partial_rank(ch,offset,ch_index,data->no_bits) - 1;	
	}
	gettimeofday(&timecheck, NULL);
        theend = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
	printf("Qnum time is %ld \n",theend-start);
	//printf("times are %f %f %f \n",trap_time/(m-1),search_time/(m-1),ret_time/(m-1));
	//printf("lookup-time is %f \n",((double)(end-begin))/CLOCKS_PER_SEC);

	*num_occ = g-f+1;
	//printf("%d \n",*num_occ);
	if(g-f+1 > 1) //this allows to retrieve only one occurrence. Comment out to retrieve all occurrences
		g = f + 1;
	unsigned long* occ = new unsigned long[g-f+1];	
	int el_f, el_g, write = 0;
	Integer * sa_blob = (Integer *) malloc(data->sa_entry_size);
	CHECK_MALLOC(sa_blob);
	el_f = f / data->sa_sample_rate;
	el_g = g / data->sa_sample_rate;
	//begin = clock();
	gettimeofday(&timecheck, NULL);
        start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
	if(el_g - el_f > 0)
	{
		pir_trapdoor(trapdoor,el_f,data->sa_depth,b,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,b);
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_f,data->sa_entry_size,res);	
		for(int j=f%data->sa_sample_rate;j< data->sa_sample_rate;j++)
		{	
			occ[write] = sa_blob[j];
			write++;
		}
		for(int i=el_f+1;i<el_g;i++)
		{
			pir_trapdoor(trapdoor,i,data->sa_depth,b,key->pk,key->hr);
			pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,b);
			pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,i,data->sa_entry_size,res);	
			for(int j=0;j<data->sa_sample_rate;j++)
			{
				occ[write] = sa_blob[j];
				write++;
			}
		}
		pir_trapdoor(trapdoor,el_g,data->sa_depth,b,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,b);
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_g,data->sa_entry_size,res);	
		for(int j=0; j <= g%data->sa_sample_rate;j++)
		{
			occ[write] = sa_blob[j];
			write++;
		}
	}
	else
	{
		pir_trapdoor(trapdoor,el_f,data->sa_depth,b,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,b);
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_f,data->sa_entry_size,res);	
		for(int j=f % data->sa_sample_rate;j<= g % data->sa_sample_rate;j++)
		{	
			occ[write] = sa_blob[j];
			write++;
		}
	}
	//end = clock();
	gettimeofday(&timecheck, NULL);
        theend = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
	printf("Qocc time is %ld \n",theend - start);
	//printf("occ-time is %f \n",((double)(end-begin))/CLOCKS_PER_SEC);

	for(int i=0;i<max_depth*b;i++)
		mpz_clear(trapdoor[i]);
	free(trapdoor);
//	free(blob);
//	free(sa_blob);
	mpz_clear(res);	
	//return occ;
	*priv_occ = occ;
}


void* map_file(const char *filename, off_t *length)
{
	int fd0;
	void *ptr;

	fd0 = open(filename, O_RDONLY);

	if(fd0 == -1)
		return NULL;

	off_t fd_size = lseek(fd0, -1, SEEK_END);
	*length = fd_size;

	ptr = mmap(NULL, fd_size, PROT_READ, MAP_SHARED, fd0, 0);

	close(fd0);

	return ptr;
}


unsigned long parse_proc_info()
{
	FILE *fd = fopen("/proc/self/statm","r");
	char *q = (char*) malloc(20); //memory consumption is surely less than a 20 digit number
	CHECK_MALLOC(q);
	size_t qlen = 20;
	for(int i=0;i<6;i++) //data + stack is the 6-th field in statm
		getdelim(&q,&qlen,32,fd);
	unsigned long mem_consumption = strtol(q,NULL,10);
	free(q);
	return mem_consumption;
}

void measure_mem_consumption(unsigned long * max)
{
	while(1)
	{
		sleep(10);
		unsigned long mem = parse_proc_info();
		if(mem > *max)
		{
			*max = mem;
		}
	}
}

int main(int argc,char **argv)
{
	if(argc < 5)
	{
		printf("input parameters missing \n");
		exit(-1);
	}	
	b = atoi(argv[3]);
	off_t length = strlen(argv[2]);
	uint64_t sample_rate = 240;
	charset * text = (charset *) map_file(argv[2],&length);
	if (text == NULL)
	{
		printf("input file does not exist \n");
		exit(-1);
	}	
	//charset * text = (charset*) argv[2];
	Integer *C = new Integer[alphabet_size + 1];
	bucket_index<charset,Integer>(text, length, alphabet_size, C);
	Integer* suffix_array = new Integer[length+1];
	sais<charset,Integer>(text,length,alphabet_size,suffix_array);
	charset* bwt = new charset[length+1];
	Integer terminator;
	build_bwt<charset,Integer>(text,suffix_array,bwt,length,&terminator);

	munmap(text, length);
	int no_bits = 1;
	while ((1 << no_bits) < alphabet_size + 1)
		no_bits++;
	int compressed_window = sample_rate * no_bits;
	int window_text_size = (compressed_window / 16) + (compressed_window % 16 == 0 ? 0 : 1);
	int no_samples = ((length+1) / sample_rate) + ((length+1) % sample_rate == 0 ? 0 : 1);
	std::size_t sample_size = sizeof(std::uint16_t) * window_text_size + sizeof(std::int32_t) * alphabet_size;
	std::size_t blob_size = no_samples * sample_size;
	std::uint8_t *blob = (std::uint8_t*) malloc(blob_size);
	CHECK_MALLOC(blob);
	sampled_bwt<charset, Integer>(bwt, length, alphabet_size, terminator, sample_rate, no_bits, sample_size, blob);
	
	key_tuple* key = (key_tuple*) malloc(sizeof(key_tuple));
	CHECK_MALLOC(key);
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}
	fread(key->enc_key,sizeof(unsigned char),BLOCK_SIZE,fp);

	dataset* data = (dataset*) malloc(sizeof(dataset));
	CHECK_MALLOC(data);
	data->sample_rate = sample_rate;
	data->sa_sample_rate = sa_sample_rate;
	data->sample_size = sample_size;
	data->no_bits = no_bits;
	data->bwt_len = no_samples;
	data->sa_len = ((length+1) / data->sa_sample_rate) + ((length+1) % data->sa_sample_rate == 0 ? 0 : 1);

	unsigned int l=0;
	unsigned int power = 1;
	while (power < data->bwt_len && power < data->sa_len)
	{
		power *= b;
		l++;
	}
	if(power >= data->bwt_len)
	{
		data->bwt_depth = l;
		while (power < data->sa_len)
		{
			power *= b;
			l++;
		}
		data->sa_depth = l;
	}
	else
	{
		data->sa_depth = l;
		while (power < data->bwt_len)
		{
			power *= b;
			l++;
		}
		data->bwt_depth = l;
	}
	int max_depth = data->bwt_depth;
	if(data->sa_depth > max_depth)
		max_depth = data->sa_depth;
	key->pk = djcs_init_public_key();
	key->sk = djcs_init_private_key();
    	key->hr = hcs_init_random();
	djcs_generate_key_pair(key->pk,key->sk,key->hr,max_depth,MODULUS_SIZE);

	data->BWT = (mpz_t *) malloc(sizeof(mpz_t)*no_samples);
	CHECK_MALLOC(data->BWT);
	data-> bwt_entry_size = build_sampled_bwt(key->enc_key,data->BWT,blob,sample_size,blob_size);
	data->SA = (mpz_t *) malloc(sizeof(mpz_t)*(length+1));
	CHECK_MALLOC(data->SA);
	data->sa_entry_size = build_suffix_array<Integer>(key->enc_key,data->SA,suffix_array,data->sa_sample_rate,length+1);

	FILE * fd = fopen(argv[1],"r");
	size_t qlen;
	unsigned int num_queries = atoi(argv[4]), count = 0;
	char *q;
	q = NULL;
	std::vector<std::string> vec;
	while((qlen = getline(&q,&qlen,fd)) != -1)
	{
		q[qlen-1] = 0;
		vec.push_back(std::string(q));
		count++;
		if(count >= num_queries)
			break;
	}
	fclose(fd);
	
	//charset * substring = (charset*) argv[1];
	//int qlen = strlen(argv[1]);
	//int num_queries = 10;
	unsigned long **private_occ = (unsigned long **) malloc(sizeof(unsigned long*)*num_queries);
	CHECK_MALLOC(private_occ);
	unsigned long **occ = (unsigned long **) malloc(sizeof(unsigned long*)*num_queries);
	CHECK_MALLOC(occ);
	unsigned long * num_occ = (unsigned long*) malloc(sizeof(unsigned long)*num_queries);
	CHECK_MALLOC(num_occ);
	unsigned long * private_num_occ = (unsigned long*) malloc(sizeof(unsigned long)*num_queries);
	CHECK_MALLOC(private_num_occ);
	key_tuple ** keys = (key_tuple **) malloc(sizeof(key_tuple *)*num_queries);
	CHECK_MALLOC(keys);
	std::thread * ths = new std::thread[num_queries];
	for(int i=0;i<num_queries;i++)
	{	
		qlen = strlen(vec[i % count].c_str());
		charset* substring = (charset *) vec[i % count].c_str();
	
		keys[i] = (key_tuple *) malloc(sizeof(key_tuple));
		CHECK_MALLOC(keys[i]);
		strncpy((char *) keys[i]->enc_key,(char *) key->enc_key,BLOCK_SIZE);
		keys[i]->pk = djcs_init_public_key();
		keys[i]->sk = djcs_init_private_key();
    		keys[i]->hr = hcs_init_random();
		djcs_generate_key_pair(keys[i]->pk,keys[i]->sk,keys[i]->hr,max_depth,MODULUS_SIZE);	
			
			occ[i] = query_ptx(substring,qlen,C,blob,data, suffix_array,&(num_occ[i]));
	}
	fd = fopen("mem_consumption.csv","a");
	unsigned long base_mem = parse_proc_info();
	fprintf(fd,"%d;%ld;%ld;",num_queries,length+1,base_mem);
	for(int i=0;i<num_queries;i++)
	{
		qlen = strlen(vec[i % count].c_str());
		charset * substring = (charset *) vec[i % count].c_str();
		ths[i] = std::thread(query,substring,qlen,C,data,max_depth,keys[i],&(private_num_occ[i]),&(private_occ[i]));
	}
	unsigned long max_mem = base_mem;
	std::thread th = std::thread(measure_mem_consumption,&max_mem);
	for(int i=0;i<num_queries;i++)
	{
		ths[i].join();
	}	
	fprintf(fd,"%ld\n",max_mem);
	fclose(fd);
	th.detach();
	for(int i=0;i<num_queries;i++)
	{
		printf("\n number of occurrences (private):%ld \t number of occurrences (non-private):%ld \n",private_num_occ[i],num_occ[i]);
		fflush(stdout);
		printf("position (private): ");
		for(int j=0;j<MIN(private_num_occ[i],1);j++)
			printf("%ld \t",private_occ[i][j]);
		printf("position (non-private): ");
		for(int j=0;j<MIN(num_occ[i],1);j++)
			printf("%ld \n",occ[i][j]);
		delete[] occ[i];	
		delete[] private_occ[i];
		djcs_cleanup(keys[i]->pk,keys[i]->sk,keys[i]->hr);
		free(keys[i]);
	}

cleanup:
	for(int i=0;i<data->bwt_len;i++)
		mpz_clear(data->BWT[i]);
	for(int i=0;i<data->sa_len;i++)
		mpz_clear(data->SA[i]);
	free(data->BWT);
	free(data->SA);
	free(data);
	djcs_cleanup(key->pk,key->sk,key->hr);
	free(key);
	free(keys);
	free(occ);
	free(num_occ);
	free(private_num_occ);
	free(private_occ);
	delete[] suffix_array;
	delete[] bwt;
	delete[] ths;
	return 0;
}
