#include "lipmaa.h"
#include "build_bwt.h"
#include <thread>

void pir_trapdoor(mpz_t* trapdoor, unsigned int index,unsigned int l,unsigned int b,djcs_public_key *pk,hcs_random *hr)
{
	//mpz_t* trapdoor = (mpz_t *) malloc(l*sizeof(mpz_t));
	mpz_t one, zero;
	mpz_inits(one,zero,NULL);
	for(int i=0;i<l;i++)
	{
		//mpz_init(trapdoor[i]);
		mpz_set_ui(one,1);
		mpz_set_ui(zero,0);
		unsigned int digit = index % b;
		for(int j=0;j<b;j++)
		{
			if(digit == j)
				djcs_encrypt_with_length(pk,hr,trapdoor[b*i+j],one,i+1);
			else
				djcs_encrypt_with_length(pk,hr,trapdoor[b*i+j],zero,i+1);
		}
		index = index / b;
	}
	mpz_clears(one,zero,NULL);
	//return trapdoor;
}

/*void dump_trapdoor(mpz_t *trapdoor, unsigned int l, unsigned int b, unsigned int modulus_size,const char*filename)
{
	unsigned long num_digits;
	unsigned char*el;
	FILE *fp = fopen(filename,"wb");
	mpz_t tmp;
	mpz_init(tmp);
	for(int i=0;i<l;i++)
	{
		num_digits = modulus_size*(i+1)/8;
		for(int j=0;j<b;j++)
		{
			mpz_set(tmp,trapdoor[i*b+j]);
			el = mpz_to_byte(tmp,num_digits);
			fwrite(el,sizeof(unsigned long),num_digits,fp);
			free(el);
		}
	}
	fclose(fp);
}*/

unsigned int dump_trapdoor(mpz_t *trapdoor, unsigned int l, unsigned int b, unsigned int modulus_size,const char*filename)
{
	unsigned long num_digits;
	unsigned char*el;
	FILE *fp = fopen(filename,"wb");
	mpz_t tmp;
	mpz_init(tmp);
	unsigned int count = 0;
	for(int i=0;i<l;i++)
	{
		num_digits = modulus_size*(i+1)/8;
		for(int j=0;j<b;j++)
		{
			mpz_set(tmp,trapdoor[i*b+j]);
			el = mpz_to_byte(tmp,num_digits);
			fwrite(el,sizeof(unsigned long),num_digits,fp);
			count += sizeof(unsigned long)*num_digits;
			free(el);
		}
	}
	fclose(fp);
	return count;
}



void verify_trapdoor(mpz_t *trapdoor,djcs_private_key *sk,unsigned long l,unsigned int b)
{
	mpz_t plain;
	mpz_init(plain);
	for(int i=0;i<l;i++)
	{
		for(int j=0;j<b;j++)
		{
			djcs_decrypt_with_length(sk,plain,trapdoor[b*i+j],i+1);
			gmp_printf("%Zd ",plain);
		}
		printf("\t");
	}
	printf("\n");
}

void recursive_pir(mpz_t res,djcs_public_key *pk, mpz_t *trapdoor, mpz_t *A,unsigned long l,unsigned int b, unsigned int begin,unsigned int end,unsigned int len)
{
	if (begin >= len)
	{
		mpz_set_ui(res,1);
		return;
	}	
	if (end - begin == 0)
	{
		mpz_set(res,A[begin]);
		return;
	}
	unsigned int bucket_size = (end - begin) / b;
	mpz_t el;
	mpz_init(el);
	recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
	begin += bucket_size +1;
	mpz_t tmp;
	mpz_init(tmp);
	mpz_powm(res,trapdoor[(l-1)*b],el,pk->n[l]);
	for(int i=1;i<b;i++)
	{
		if(begin >= len)
			break;
		recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
		begin += bucket_size +1;
		mpz_powm(tmp,trapdoor[(l-1)*b+i],el,pk->n[l]);
		mpz_mul(res,res,tmp);
		mpz_mod(res,res,pk->n[l]);
	}
	mpz_clear(el);
}

void multi_recursive_pir(mpz_t res,djcs_public_key *pk, mpz_t *trapdoor, mpz_t *A,unsigned long l,unsigned int b, unsigned int begin,unsigned int end,unsigned int len)
{
	if (end - begin == 0)
	{
		mpz_set(res,A[begin]);
	//	printf("%d %d \n",begin,l);
	//	fflush(stdout);
		return;
	}
	
	mpz_t *results = (mpz_t *) malloc(sizeof(mpz_t)*b);
	std::thread * threads = new std::thread[b]; 
	for(int i=0;i<b;i++)
		mpz_init(results[i]);
	unsigned int bucket_size = (end - begin) / b;
	for(int i=0;i<b;i++)
	{
		threads[i] = std::thread(recursive_pir,results[i],pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
		begin +=bucket_size+1;
	}
	for(int i=0;i<b;i++)
		threads[i].join();

	mpz_t tmp;
	mpz_init(tmp);
	mpz_powm(res,trapdoor[(l-1)*b],results[0],pk->n[l]);
	for(int i=1;i<b;i++)
	{
		mpz_powm(tmp,trapdoor[(l-1)*b+i],results[i],pk->n[l]);
		mpz_mul(res,res,tmp);
		mpz_mod(res,res,pk->n[l]);
	}
	for(int i=0;i<b;i++)
		mpz_clear(results[i]);
	free(results);
	delete[] threads;
	
}

void init_recursive_pir(mpz_t res,djcs_public_key *pk, mpz_t *trapdoor, mpz_t *A,unsigned long l,unsigned int b, unsigned int begin,unsigned int end,unsigned int len)
{
	if (begin >= len)
	{
		mpz_set_ui(res,1);
		return;
	}	
	
	if (end - begin == 0)
	{
		mpz_set(res,A[begin]);
		return;
	}
	
	unsigned int bucket_size = (end - begin) / b;
	mpz_t tmp,el;
	mpz_inits(tmp,el,NULL);
	if(begin + bucket_size > len)
		init_recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
	else
		multi_recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
	begin += bucket_size + 1;
	mpz_powm(res,trapdoor[(l-1)*b],el,pk->n[l]);
	for(int i=1;i<b;i++)
	{
		if(begin >= len)
			break;
		if(begin + bucket_size > len)
			init_recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
		else
			multi_recursive_pir(el,pk,trapdoor,A,l-1,b,begin,begin+bucket_size,len);
		begin +=bucket_size+1;
		mpz_powm(tmp,trapdoor[(l-1)*b+i],el,pk->n[l]);
		mpz_mul(res,res,tmp);
		mpz_mod(res,res,pk->n[l]);
	}
	mpz_clears(tmp,el,NULL);
	
}
void pir_search(mpz_t res,djcs_public_key *pk,mpz_t *trapdoor,unsigned long l,mpz_t *A,unsigned int len,unsigned int b)
{
	unsigned int power = 1,begin = 0;
	for(int i=0;i<l;i++)
		power *= b;
	if(len > power / 2) //b threads are balanced, so split the computation among them
		multi_recursive_pir(res,pk,trapdoor,A,l,b,0,power-1,len);
	else //the b threads would be too unbalanced, thus go one recursive layer down
		init_recursive_pir(res,pk,trapdoor,A,l,b,0,power-1,len);
}

void pir_retrieve(unsigned char* el, djcs_private_key *sk, unsigned char* enc_key,unsigned long l, unsigned long index, int entry_size, mpz_t ciphertext)
{
	for(;l>0;l--)
	{
		djcs_decrypt_with_length(sk,ciphertext,ciphertext,l);
	}
	unsigned int num_digits = (entry_size)/8;
	unsigned char* el_ctx = mpz_to_byte(ciphertext,(entry_size)%8 == 0 ? num_digits : num_digits +1 );
	int len;
	EVP_CIPHER_CTX* ctx = EVP_CIPHER_CTX_new();
	if(1 != EVP_DecryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,enc_key,el_ctx) )
	{
		printf("decryption initialization error \n");
		exit(-1);
	}
	if(1 != EVP_DecryptUpdate(ctx,(unsigned char *)el,&len,el_ctx+BLOCK_SIZE,entry_size-BLOCK_SIZE) )
	{
		printf("decryption failed \n");
		exit(-1);
	}
	if(1 != EVP_DecryptFinal_ex(ctx, ((unsigned char *) el) + len, &len))
	{
		printf("final decryption failed \n");
		exit(-1);
	}	
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
}

