#include "build_bwt.h"
#include "lipmaa.h"

const int alphabet_size = 5;
typedef fasta_nucleotide charset;
typedef std::uint32_t Integer;
const int sa_sample_rate = 28;

enum test {radix_tune,single_core} test_type; 
const char* filenames[3] = {"radix_tune.csv","single-core-data.csv"};

void byte_to_mpz(unsigned char* bytes,unsigned int len_bytes,mpz_t dest)
{
	mpz_set_ui(dest,0);
	unsigned long *el = (unsigned long*) bytes;
	mpz_add_ui(dest,dest,*el);
	for(int i=sizeof(unsigned long);i<len_bytes;i+=sizeof(unsigned long))
	{
		el = (unsigned long*) &(bytes[i]);
		mpz_mul_2exp(dest,dest,64);
		mpz_add_ui(dest,dest,*el);
	}
}

unsigned char* mpz_to_byte(mpz_t op, unsigned int num_digits)
{
	unsigned long* dest = (unsigned long*) malloc(sizeof(unsigned long)*num_digits);
	mpz_t tmp;
	mpz_inits(tmp, NULL);
	for(int i=num_digits;i>0;i--)
	{
		mpz_fdiv_r_2exp(tmp,op,64);
		mpz_fdiv_q_2exp(op,op,64);
		dest[i-1] = mpz_get_ui(tmp); 	
	}
	mpz_clear(tmp);
	return (unsigned char*) dest;
}



int build_sampled_bwt(key_tuple *key,mpz_t *A,std::uint8_t* ptx_array, int entry_size, int array_size)
{	
	unsigned char iv[BLOCK_SIZE];
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}

	EVP_CIPHER_CTX *ctx;
	ctx = EVP_CIPHER_CTX_new();
	if(!ctx)
	{
		printf("Wrong ctx initialization \n");
		exit(-1);
	}
	if(1 != EVP_EncryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,key->enc_key,NULL) )
	{
		printf("encryption initialization error \n");
		exit(-1);
	}
	std::uint8_t *el_ctx = (std::uint8_t*) malloc((entry_size+2)*BLOCK_SIZE);
	int ctx_len,len,j=0;
	for(unsigned int i=0;i<array_size;i+=entry_size)
	{
		fread(iv, sizeof(unsigned char), BLOCK_SIZE, fp);
		if(1 != EVP_EncryptInit_ex(ctx,NULL,NULL,NULL,iv))
		{
			printf("encryption iv setup failed \n");
			exit(-1);
		}
		memcpy(el_ctx,iv,BLOCK_SIZE);
		ctx_len = BLOCK_SIZE;
		if(1 != EVP_EncryptUpdate(ctx,el_ctx+ctx_len,&len,&(ptx_array[i]),entry_size))
		{
			printf("encryption failed \n");
			exit(-1);
		}
		ctx_len += len; 
		if(1 != EVP_EncryptFinal_ex(ctx,el_ctx+ctx_len,&len))
		{
			printf("padding failed \n");
			exit(-1);
		}
		ctx_len += len;
		mpz_init(A[j]);
		byte_to_mpz(el_ctx,ctx_len,A[j]);	
		//gmp_printf(" element %d is %Zd \n",j,A[j]);
		j++;
	}
	fclose(fp);
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
	return ctx_len;
}


template <typename Int>
int build_suffix_array(key_tuple *key,mpz_t *A,Int* ptx_array,int sample_rate, int N)
{
	unsigned char iv[BLOCK_SIZE];
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}

	EVP_CIPHER_CTX *ctx;
	ctx = EVP_CIPHER_CTX_new();
	if(!ctx)
	{
		printf("Wrong ctx initialization \n");
		exit(-1);
	}
	if(1 != EVP_EncryptInit_ex(ctx,EVP_aes_128_ctr(),NULL,key->enc_key,NULL) )
	{
		printf("encryption initialization error \n");
		exit(-1);
	}
	unsigned char *element,*el_ctx = (unsigned char *) malloc(sizeof(Int)*sample_rate + BLOCK_SIZE);
	int ctx_len,len,j=0;
	for(unsigned long i=0;i<N;i+=sample_rate)
	{
		//unsigned char * element = malloc(sizeof(unsigned long));
		//memcpy(element,&i,sizeof(unsigned long));
		element = (unsigned char *) &(ptx_array[i]);
		fread(iv, sizeof(unsigned char), BLOCK_SIZE, fp);
		if(1 != EVP_EncryptInit_ex(ctx,NULL,NULL,NULL,iv))
		{
			printf("encryption iv setup failed \n");
			exit(-1);
		}
		memcpy(el_ctx,iv,BLOCK_SIZE);
		ctx_len = BLOCK_SIZE;
		if(1 != EVP_EncryptUpdate(ctx,el_ctx+ctx_len,&len,element,sizeof(Int)*sample_rate))
		{
			printf("encryption failed \n");
			exit(-1);
		}
		ctx_len += len; 
		if(1 != EVP_EncryptFinal_ex(ctx,el_ctx+ctx_len,&len))
		{
			printf("padding failed \n");
			exit(-1);
		}
		ctx_len += len;	
		mpz_init(A[j]);
		byte_to_mpz(el_ctx,ctx_len,A[j]);
		j++;
	}
	fclose(fp);
	EVP_CIPHER_CTX_free(ctx);
	free(el_ctx);
	return ctx_len;
}


unsigned long* query_ptx(charset* substring, int m, Integer* ch_count,std::uint8_t* BWT, dataset* data, Integer* SA, unsigned long* num_occ)
{
	/*Integer ch_index = ((Integer) substring[m-1]); //+1 because the first row of the bwt is due to the terminator
	unsigned long f = ch_count[ch_index] - 1; //-1 needs to be added because indexes of the array starts from 0
	unsigned long g = ch_count[ch_index+1] - 1;
	for(int i=m-2;i>=0;i--)
	{
		ch_index = ((Integer) substring[i]) + 1; //+1 because the first row of the bwt is due to the terminator
		f = ch_index*length + f;
		f = BWT[f] - 1;
		g = ch_index*length + g;
		g = BWT[g] - 1;
	}*/
	Integer start, end;
	vbwt_query<Integer,charset>(BWT,ch_count,data->sample_rate,data->sample_size,data->no_bits,alphabet_size,substring,m,&start,&end);
	
	*num_occ = end-start+1;
	if(*num_occ != 0)
	{
		unsigned long* occ = new unsigned long[*num_occ];
		for(int i=start;i<end+1;i++)
			occ[i-start] = SA[i];
		return occ;
	}
	return NULL;
}

unsigned long* query(charset* substring, int m, Integer* ch_count,dataset* data, int max_depth, key_tuple* key, unsigned long* num_occ, unsigned int doc_size)
{
	Integer ch_index = ((Integer) substring[m-1]);
	unsigned long f = ch_count[ch_index]; //-1 needs to be added because indexes of the array starts from 0
	unsigned long g = ch_count[ch_index+1] - 1;
	unsigned long offset;
	std::uint32_t* blob = (std::uint32_t*) malloc(data->sample_size);
	std::uint16_t* ch;
	mpz_t * trapdoor = (mpz_t*) malloc(sizeof(mpz_t)*max_depth*data->b_bwt);
	for(int i=0;i<max_depth*data->b_bwt;i++)
		mpz_init(trapdoor[i]);
	mpz_t res;
	mpz_init(res);
	//clock_t begin, end;
	//double trap_time = 0, search_time= 0, ret_time = 0;
	//begin = clock();
	long start, end;
	long client_time = 0, server_time = 0;
	struct timeval timecheck;
	char line[100],str[100];
	unsigned int max_comm_cost = 0,comm_cost,str_count;
	if(test_type != radix_tune) 
		str_count = sprintf(line,"%d;",doc_size);
	else
		str_count = sprintf(line,"%d;",data->b_bwt);
	for(int i=m-2;i>=0;i--)
	{
		//printf("start is % d end is %d \n",f,g);
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		ch_index = ((Integer) substring[i]);
		offset = f % data->sample_rate;
		f = f / data->sample_rate;
		//begin = clock();
		pir_trapdoor(trapdoor,f,data->bwt_depth,data->b_bwt,key->pk,key->hr);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		client_time += end - start;
		
		//end = clock();
		//trap_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		comm_cost = dump_trapdoor(trapdoor,data->bwt_depth,data->b_bwt,MODULUS_SIZE,"trapdoor");
		max_comm_cost = comm_cost > max_comm_cost ? comm_cost : max_comm_cost; 
		//begin = clock();
	
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_search(res,key->pk,trapdoor,data->bwt_depth,data->BWT,data->bwt_len,data->b_bwt);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		server_time += end - start;
		
		//end = clock();
		//search_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		//begin = clock();
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_retrieve((unsigned char*) blob,key->sk,key->enc_key,data->bwt_depth,f,data->bwt_entry_size,res);
		//end = clock();
		ch = (std::uint16_t *) (blob + alphabet_size);
		f = ch_count[ch_index];
		f += blob[ch_index] + partial_rank(ch,offset,ch_index,data->no_bits);
		//ret_time += ((double) (end-begin))/CLOCKS_PER_SEC;
		offset = g % data->sample_rate + 1;
		g = g / data->sample_rate;
		pir_trapdoor(trapdoor,g,data->bwt_depth,data->b_bwt,key->pk,key->hr);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		client_time += end - start;
		
		comm_cost = dump_trapdoor(trapdoor,data->bwt_depth,data->b_bwt,MODULUS_SIZE,"trapdoor");
		max_comm_cost = comm_cost > max_comm_cost ? comm_cost : max_comm_cost;
		
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_search(res,key->pk,trapdoor,data->bwt_depth,data->BWT,data->bwt_len,data->b_bwt);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		server_time += end - start;
		
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_retrieve((unsigned char*) blob,key->sk,key->enc_key,data->bwt_depth,g,data->bwt_entry_size,res);
		ch = (std::uint16_t *) (blob + alphabet_size);
		g = ch_count[ch_index];
		g += blob[ch_index] + partial_rank(ch,offset,ch_index,data->no_bits) - 1;	
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		client_time += end - start;
		
	}
	printf("Qnum times are %ld %ld \n",client_time,server_time);
	str_count += sprintf(str,"%ld;%ld;%d;",client_time,server_time,max_comm_cost);
	strcat(line,str);
	//printf("times are %f %f %f \n",trap_time/(m-1),search_time/(m-1),ret_time/(m-1));
	//printf("lookup-time is %f \n",((double)(end-begin))/CLOCKS_PER_SEC);

	*num_occ = g-f+1;
	//printf("%d \n",*num_occ);
	if(g-f+1 > 1) //this allows to retrieve only one occurrence. Comment out to retrieve all occurrences
		g = f;
	unsigned long* occ = new unsigned long[*num_occ];	
	for(int i=0;i<max_depth*data->b_bwt;i++)
		mpz_clear(trapdoor[i]);
	free(trapdoor);
	trapdoor = (mpz_t*) malloc(sizeof(mpz_t)*max_depth*data->b_sa);
	for(int i=0;i<max_depth*data->b_sa;i++)
		mpz_init(trapdoor[i]);
	
	int el_f, el_g, write = 0;
	Integer * sa_blob = (Integer *) malloc(data->sa_entry_size);
	el_f = f / data->sa_sample_rate;
	el_g = g / data->sa_sample_rate;
	//begin = clock();
	client_time = 0;
	server_time = 0;
	max_comm_cost = 0;
	if(el_g - el_f > 0)
	{
		pir_trapdoor(trapdoor,el_f,data->sa_depth,data->b_sa,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,data->b_sa);
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_f,data->sa_entry_size,res);	
		for(int j=f%data->sa_sample_rate;j< data->sa_sample_rate;j++)
		{	
			occ[write] = sa_blob[j];
			write++;
		}
		for(int i=el_f+1;i<el_g;i++)
		{
			pir_trapdoor(trapdoor,i,data->sa_depth,data->b_sa,key->pk,key->hr);
			pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,data->b_sa);
			pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,i,data->sa_entry_size,res);	
			for(int j=0;j<data->sa_sample_rate;j++)
			{
				occ[write] = sa_blob[j];
				write++;
			}
		}
		pir_trapdoor(trapdoor,el_g,data->sa_depth,data->b_sa,key->pk,key->hr);
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,data->b_sa);
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_g,data->sa_entry_size,res);	
		for(int j=0; j <= g%data->sa_sample_rate;j++)
		{
			occ[write] = sa_blob[j];
			write++;
		}
	}
	else
	{
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_trapdoor(trapdoor,el_f,data->sa_depth,data->b_sa,key->pk,key->hr);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		client_time += end - start;
	
		comm_cost = dump_trapdoor(trapdoor,data->sa_depth,data->b_sa,MODULUS_SIZE,"trapdoor");
		max_comm_cost = comm_cost > max_comm_cost ? comm_cost : max_comm_cost;
		
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_search(res,key->pk,trapdoor,data->sa_depth,data->SA,data->sa_len,data->b_sa);
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		server_time += end - start;
		
		gettimeofday(&timecheck, NULL);
        	start = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		pir_retrieve((unsigned char*) sa_blob,key->sk,key->enc_key,data->sa_depth,el_f,data->sa_entry_size,res);	
		for(int j=f % data->sa_sample_rate;j<= g % data->sa_sample_rate;j++)
		{	
			occ[write] = sa_blob[j];
			write++;
		}
		gettimeofday(&timecheck, NULL);
        	end = (long)timecheck.tv_sec * 1000 + (long)timecheck.tv_usec / 1000;
		server_time += end - start;
	}
	//end = clock();
	printf("Qocc times are %ld %ld \n",client_time, server_time);
	str_count += sprintf(str,"%ld;%ld;%d\n",client_time,server_time,max_comm_cost/(2*m));	
	strcat(line,str);

	FILE * fd = fopen(filenames[test_type],"a");
	fwrite(line,1,str_count,fd);
	fclose(fd);
	//printf("occ-time is %f \n",((double)(end-begin))/CLOCKS_PER_SEC);

	for(int i=0;i<max_depth*data->b_sa;i++)
		mpz_clear(trapdoor[i]);
	free(trapdoor);
	free(blob);
	free(sa_blob);
	mpz_clear(res);	
	return occ;
}


void* map_file(const char *filename, off_t *length)
{
	int fd0;
	void *ptr;

	fd0 = open(filename, O_RDONLY);

	if(fd0 == -1)
		return NULL;

	off_t fd_size = lseek(fd0, -1, SEEK_END);
	*length = fd_size;

	ptr = mmap(NULL, fd_size, PROT_READ, MAP_SHARED, fd0, 0);

	close(fd0);

	return ptr;
}

#define UPPER_B 28
#define LOWER_B 12

unsigned int find_optimal_b(unsigned int len, unsigned int *max_depth)
{
	unsigned int power = 1, depth = 0;
	while(power < len)
	{
		power *= UPPER_B;
		depth++;
	}
	unsigned int max_delta = power - len;
	unsigned int opt_b = UPPER_B;
	*max_depth = depth;
	for(int b=UPPER_B-1;b > LOWER_B;b--)
	{
		power = 1; 
		depth = 0;
		while(power < len)
		{
			power *= b;
			depth++;
		}
		//if(depth > *max_depth)
		//	return opt_b;
		if(power - len < max_delta)
		{
			max_delta = power - len;
			opt_b = b;
			*max_depth = depth;
		}
	}
	return opt_b;
}

int main(int argc,char **argv)
{
	if(argc < 5)
	{
		printf("input parameters missing \n");
		exit(-1);
	}
	test_type = (test) atoi(argv[4]);
	off_t length = strlen(argv[2]);
	uint64_t sample_rate = 240;
	charset * text = (charset *) map_file(argv[2],&length);
	//charset * text = (charset*) argv[2];
	if (text == NULL)
	{
		printf("input file does not exist \n");
		exit(-1);
	}
	Integer *C = new Integer[alphabet_size + 1];
	bucket_index<charset,Integer>(text, length, alphabet_size, C);
	/*for(int i=0;i<alphabet_size+1;i++)
		printf("%d ",C[i]);
	printf("\n");*/
	Integer* suffix_array = new Integer[length+1];
	sais<charset,Integer>(text,length,alphabet_size,suffix_array);
	charset* bwt = new charset[length+1];
	Integer terminator;
	build_bwt<charset,Integer>(text,suffix_array,bwt,length,&terminator);
	munmap(text, length);
	int no_bits = 1;
	while ((1 << no_bits) < alphabet_size + 1)
		no_bits++;
	int compressed_window = sample_rate * no_bits;
	int window_text_size = (compressed_window / 16) + (compressed_window % 16 == 0 ? 0 : 1);
	int no_samples = ((length+1) / sample_rate) + ((length+1) % sample_rate == 0 ? 0 : 1);
	std::size_t sample_size = sizeof(std::uint16_t) * window_text_size + sizeof(std::int32_t) * alphabet_size;
	std::size_t blob_size = no_samples * sample_size;
	std::uint8_t *blob = (std::uint8_t*) malloc(blob_size);
	sampled_bwt<charset, Integer>(bwt, length, alphabet_size, terminator, sample_rate, no_bits, sample_size, blob);
	

	key_tuple* key = (key_tuple*) malloc(sizeof(key_tuple));
	FILE *fp = fopen(RANDOM_SOURCE, "rb");
	if (fp == NULL) {
		printf("Cannot open random source! (file=%s, mode=rb)\n", RANDOM_SOURCE);
		exit(4);
	}
	fread(key->enc_key,sizeof(unsigned char),BLOCK_SIZE,fp);
	
	dataset* data = (dataset*) malloc(sizeof(dataset));
	data->sample_rate = sample_rate;
	data->sa_sample_rate = sa_sample_rate;
	data->sample_size = sample_size;
	data->no_bits = no_bits;
	data->bwt_len = no_samples;
	data->sa_len = ((length+1) / data->sa_sample_rate) + ((length+1) % data->sa_sample_rate == 0 ? 0 : 1);
	
	int b = atoi(argv[3]);
	data->b_bwt = b;
	data->b_sa = b;
	unsigned int l=0;
	unsigned int power = 1;
	while (power < data->bwt_len && power < data->sa_len)
	{
		power *= b;
		l++;
	}
	if(power >= data->bwt_len)
	{
		data->bwt_depth = l;
		while (power < data->sa_len)
		{
			power *= b;
			l++;
		}
		data->sa_depth = l;
	}
	else
	{
		data->sa_depth = l;
		while (power < data->bwt_len)
		{
			power *= b;
			l++;
		}
		data->bwt_depth = l;
	}
	int max_depth = data->bwt_depth;
	if(data->sa_depth > max_depth)
		max_depth = data->sa_depth;
	key->pk = djcs_init_public_key();
	key->sk = djcs_init_private_key();
    	key->hr = hcs_init_random();
	djcs_generate_key_pair(key->pk,key->sk,key->hr,max_depth,MODULUS_SIZE);
	data->BWT = (mpz_t *) malloc(sizeof(mpz_t)*no_samples);
	data-> bwt_entry_size = build_sampled_bwt(key,data->BWT,blob,sample_size,blob_size);
	data->SA = (mpz_t *) malloc(sizeof(mpz_t)*(length+1));
	data->sa_entry_size = build_suffix_array<Integer>(key,data->SA,suffix_array,data->sa_sample_rate,length+1);

	charset * substring = (charset *) argv[1];
	unsigned long num_occ,private_num_occ;
	

	unsigned long* private_occ = query(substring,strlen(argv[1]),C,data,max_depth,key,&private_num_occ,length+1);

		unsigned long* occ = query_ptx(substring,strlen(argv[1]),C,blob,data, suffix_array,&num_occ);
		/*for(int i=0;i<length+1;i++)
			printf("%d ",suffix_array[i]);	
		*/
		printf("\n number of occurrences (private):%ld \t number of occurrences (non-private):%ld \n",private_num_occ,num_occ);
		printf("position (private): ");
		for(int i=0;i<MIN(private_num_occ,1);i++)
			printf("%ld \t",private_occ[i]);
		printf("position (non-private): ");
		for(int i=0;i<MIN(num_occ,1);i++)
			printf("%ld \n",occ[i]);
		
	delete[] occ;	
	delete[] private_occ;
	for(int i=0;i<data->bwt_len;i++)
		mpz_clear(data->BWT[i]);
	for(int i=0;i<data->sa_len;i++)
		mpz_clear(data->SA[i]);
	free(data->BWT);
	free(data->SA);
	free(data);
	delete[] suffix_array;
	delete[] bwt;
	return 0;
}
